﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;
using HeuristicLab.Common;
using HeuristicLab.Core.Views;
using HeuristicLab.MainForm;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Encodings.SymbolicExpressionTreeEncoding.Views {
  [View("Symbolic Expression Grammar Editor")]
  [Content(typeof(ISymbolicExpressionGrammar), false)]
  public partial class SymbolicExpressionGrammarEditorView : NamedItemView {
    public SymbolicExpressionGrammarEditorView() {
      InitializeComponent();
    }

    public override bool ReadOnly {
      get {
        if ((Content != null) && Content.ReadOnly) return true;
        return base.ReadOnly;
      }
      set {
        if ((Content != null) && Content.ReadOnly) base.ReadOnly = true;
        else base.ReadOnly = value;
      }
    }

    public new ISymbolicExpressionGrammar Content {
      get { return (ISymbolicExpressionGrammar)base.Content; }
      set { base.Content = value; }
    }

    #region events
    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      Content.ReadOnlyChanged += new System.EventHandler(Content_ReadOnlyChanged);
      Content.Changed += new System.EventHandler(Content_Changed);
    }
    protected override void DeregisterContentEvents() {
      Content.ReadOnlyChanged -= new System.EventHandler(Content_ReadOnlyChanged);
      Content.Changed -= new System.EventHandler(Content_Changed);
      base.DeregisterContentEvents();
    }

    private void Content_ReadOnlyChanged(object sender, EventArgs e) {
      ReadOnly = Content.ReadOnly;
    }

    private void Content_Changed(object sender, EventArgs e) {
      ISymbol symbol = null;
      if (symbolsTreeView.SelectedNode != null)
        symbol = (ISymbol)symbolsTreeView.SelectedNode.Tag;

      OnContentChanged();

      if (symbol != null)
        symbolsTreeView.SelectedNode = IterateTreeNodes().Where(n => n.Tag == symbol).FirstOrDefault();
    }
    #endregion

    protected override void SetEnabledStateOfControls() {
      base.SetEnabledStateOfControls();
    }

    protected override void OnContentChanged() {
      base.OnContentChanged();
      if (Content != null) {
        RebuildSymbolsTreeView();
        symbolicExpressionTreeChart.Tree = null;
        symbolViewHost.Content = null;
      } else {
        symbolsTreeView.Nodes.Clear();
        symbolicExpressionTreeChart.Tree = null;
        symbolViewHost.Content = null;
      }
    }

    private void RebuildSymbolsTreeView() {
      symbolsTreeView.Nodes.Clear();
      var symbols = Content.Symbols.ToList();
      var groupSymbols = symbols.OfType<GroupSymbol>().ToList();
      var topLevelSymbols = Content.Symbols.Where(s => !groupSymbols.Any(g => g.Symbols.Contains(s)));
      AddChildTreeNodes(symbolsTreeView.Nodes, topLevelSymbols);

      RebuildImageList();
      symbolsTreeView.ExpandAll();
    }

    private void AddChildTreeNodes(TreeNodeCollection collection, IEnumerable<ISymbol> symbols) {
      foreach (ISymbol symbol in symbols) {
        var node = new TreeNode();
        node.Name = symbol.Name;
        node.Text = symbol.Name;
        node.Tag = symbol;
        node.Checked = symbol.Enabled;
        collection.Add(node);
        var groupSymbol = symbol as GroupSymbol;
        if (groupSymbol != null) AddChildTreeNodes(node.Nodes, groupSymbol.Symbols);
      }
    }

    private void symbolsTreeView_AfterSelect(object sender, TreeViewEventArgs e) {
      var symbol = (ISymbol)e.Node.Tag;
      if (!(symbol is DummySymbol)) {
        symbolViewHost.Content = symbol;
        symbolicExpressionTreeChart.Tree = BuildAllowedChildSymbolsTree(symbol);
      } else {
        symbolViewHost.Content = null;
        symbolicExpressionTreeChart.Tree = null;
      }
    }

    private void symbolsTreeView_AfterCheck(object sender, TreeViewEventArgs e) {
      if (e.Action != TreeViewAction.Unknown) {
        DeregisterContentEvents();
        var symbol = (ISymbol)e.Node.Tag;
        symbol.Enabled = e.Node.Checked;
        RegisterContentEvents();
        Content_Changed(Content, EventArgs.Empty);
      }
    }

    private ISymbolicExpressionTree BuildAllowedChildSymbolsTree(ISymbol symbol) {
      var tree = new SymbolicExpressionTree(new SymbolicExpressionTreeNode(symbol));

      if (Content.GetMaximumSubtreeCount(symbol) > 0) {
        for (int i = 0; i < Content.GetMaximumSubtreeCount(symbol); i++) {
          var node = new DummySymbol("Subtree " + i).CreateTreeNode();
          foreach (var childSymbol in Content.GetAllowedChildSymbols(symbol, i)) {
            node.AddSubtree(new SymbolicExpressionTreeNode(childSymbol));
          }
          tree.Root.AddSubtree(node);
        }
      }
      return tree;
    }


    #region helpers
    private IEnumerable<TreeNode> IterateTreeNodes(TreeNode node = null) {
      TreeNodeCollection nodes;
      if (node == null)
        nodes = symbolsTreeView.Nodes;
      else {
        nodes = node.Nodes;
        yield return node;
      }

      foreach (var childNode in nodes.OfType<TreeNode>())
        foreach (var n in IterateTreeNodes(childNode))
          yield return n;
    }

    protected virtual void RebuildImageList() {
      symbolsTreeView.ImageList.Images.Clear();
      foreach (TreeNode treeNode in IterateTreeNodes()) {
        var symbol = (ISymbol)treeNode.Tag;
        symbolsTreeView.ImageList.Images.Add(symbol == null ? HeuristicLab.Common.Resources.VSImageLibrary.Nothing : symbol.ItemImage);
        treeNode.ImageIndex = symbolsTreeView.ImageList.Images.Count - 1;
      }
    }
  }
    #endregion

  [NonDiscoverableType]
  internal class DummySymbol : Symbol {
    public DummySymbol(DummySymbol original, Cloner cloner) : base(original, cloner) { }
    public DummySymbol(string name) : base(name, "DummySymbol for views") { }
    public override IDeepCloneable Clone(Cloner cloner) { return new DummySymbol(this, cloner); }
  }
}
